import multiprocessing
import sys


def preload_modules(context: multiprocessing.context.BaseContext) -> None:
    all_loaded_modules = sys.modules.keys()
    preload = [
        loaded_module for loaded_module in all_loaded_modules
        if loaded_module.split('.')[0] in (
            'smac',
            'autosklearn',
            'numpy',
            'scipy',
            'pandas',
            'pynisher',
            'sklearn',
            'ConfigSpace',
        ) and 'logging' not in loaded_module
    ]
    context.set_forkserver_preload(preload)



from typing import Optional, Union

from ConfigSpace.configuration_space import ConfigurationSpace

import numpy as np

from sklearn.base import BaseEstimator
from sklearn.exceptions import NotFittedError

from autosklearn.pipeline.base import DATASET_PROPERTIES_TYPE, PIPELINE_DATA_DTYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm


class Rescaling(object):
    # Rescaling does not support fit_transform (as of 0.19.1)!
    def __init__(
        self,
        random_state: Optional[Union[int, np.random.RandomState]] = None
    ) -> None:
        self.preprocessor: Optional[BaseEstimator] = None

    def fit(self, X: PIPELINE_DATA_DTYPE, y: Optional[PIPELINE_DATA_DTYPE] = None
            ) -> 'AutoSklearnPreprocessingAlgorithm':
        if self.preprocessor is None:
            raise NotFittedError()
        self.preprocessor.fit(X)
        return self

    def transform(self, X: PIPELINE_DATA_DTYPE) -> PIPELINE_DATA_DTYPE:
        if self.preprocessor is None:
            raise NotImplementedError()
        return self.preprocessor.transform(X)

    @staticmethod
    def get_hyperparameter_search_space(dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None
                                        ) -> ConfigurationSpace:
        cs = ConfigurationSpace()
        return cs



from typing import Dict, Optional, Tuple, Union

import numpy as np

from autosklearn.pipeline.base import DATASET_PROPERTIES_TYPE
from autosklearn.pipeline.constants import DENSE, UNSIGNED_DATA, INPUT, SPARSE
from autosklearn.pipeline.components.data_preprocessing.rescaling.abstract_rescaling \
    import Rescaling
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm


class NormalizerComponent(Rescaling, AutoSklearnPreprocessingAlgorithm):
    def __init__(
        self,
        random_state: Optional[Union[int, np.random.RandomState]] = None
    ) -> None:
        # Use custom implementation because sklearn implementation cannot
        # handle float32 input matrix
        from sklearn.preprocessing import Normalizer
        self.preprocessor = Normalizer(copy=False)

    @staticmethod
    def get_properties(dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None
                       ) -> Dict[str, Optional[Union[str, int, bool, Tuple]]]:
        return {'shortname': 'Normalizer',
                'name': 'Normalizer',
                'handles_missing_values': False,
                'handles_nominal_values': False,
                'handles_numerical_features': True,
                'prefers_data_scaled': False,
                'prefers_data_normalized': False,
                'handles_regression': True,
                'handles_classification': True,
                'handles_multiclass': True,
                'handles_multilabel': True,
                'handles_multioutput': True,
                'is_deterministic': True,
                # TODO find out if this is right!
                'handles_sparse': True,
                'handles_dense': True,
                'input': (SPARSE, DENSE, UNSIGNED_DATA),
                'output': (INPUT,),
                'preferred_dtype': None}



# -*- encoding: utf-8 -*-
from typing import Dict, Optional, Union, cast

import numpy as np

import pandas as pd

from scipy import sparse

from autosklearn.constants import (
    BINARY_CLASSIFICATION,
    MULTICLASS_CLASSIFICATION,
    MULTILABEL_CLASSIFICATION,
    MULTIOUTPUT_REGRESSION,
    REGRESSION,
)
from autosklearn.data.abstract_data_manager import AbstractDataManager
from autosklearn.data.validation import (
    SUPPORTED_FEAT_TYPES,
    SUPPORTED_TARGET_TYPES,
)


class XYDataManager(AbstractDataManager):

    def __init__(
        self,
        X: SUPPORTED_FEAT_TYPES,
        y: SUPPORTED_TARGET_TYPES,
        X_test: Optional[SUPPORTED_FEAT_TYPES],
        y_test: Optional[SUPPORTED_TARGET_TYPES],
        task: int,
        feat_type: Dict[Union[str, int], str],
        dataset_name: str
    ):
        super(XYDataManager, self).__init__(dataset_name)

        self.info['task'] = task
        if sparse.issparse(X):
            self.info['is_sparse'] = 1
            self.info['has_missing'] = np.all(np.isfinite(cast(sparse.csr_matrix, X).data))
        else:
            self.info['is_sparse'] = 0
            if hasattr(X, 'iloc'):
                self.info['has_missing'] = cast(pd.DataFrame, X).isnull().values.any()
            else:
                self.info['has_missing'] = np.all(np.isfinite(X))

        label_num = {
            REGRESSION: 1,
            BINARY_CLASSIFICATION: 2,
            MULTIOUTPUT_REGRESSION: np.shape(y)[-1],
            MULTICLASS_CLASSIFICATION: len(np.unique(y)),
            MULTILABEL_CLASSIFICATION: np.shape(y)[-1]
        }

        self.info['label_num'] = label_num[task]

        self.data['X_train'] = X
        self.data['Y_train'] = y
        if X_test is not None:
            self.data['X_test'] = X_test
        if y_test is not None:
            self.data['Y_test'] = y_test

        if isinstance(feat_type, dict):
            self.feat_type = feat_type
        else:
            raise ValueError("Unsupported feat_type provided. We expect the user to "
                             "provide a Dict[str, str] mapping from column to categorical/ "
                             "numerical.")

        # TODO: try to guess task type!

        if len(np.shape(y)) > 2:
            raise ValueError('y must not have more than two dimensions, '
                             'but has %d.' % len(np.shape(y)))

        if np.shape(X)[0] != np.shape(y)[0]:
            raise ValueError('X and y must have the same number of '
                             'datapoints, but have %d and %d.' % (np.shape(X)[0],
                                                                  np.shape(y)[0]))



from typing import Dict, Optional, Tuple, Union


from ConfigSpace.configuration_space import ConfigurationSpace

import numpy as np
from autosklearn.pipeline.base import DATASET_PROPERTIES_TYPE, PIPELINE_DATA_DTYPE
from autosklearn.pipeline.components.base import \
    AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.constants import DENSE, SPARSE, UNSIGNED_DATA, INPUT


class NoCoalescence(AutoSklearnPreprocessingAlgorithm):
    def __init__(
        self,
        random_state: Optional[Union[int, np.random.RandomState]] = None
    ) -> None:
        pass

    def fit(self, X: np.array, y: Optional[PIPELINE_DATA_DTYPE] = None
            ) -> PIPELINE_DATA_DTYPE:
        self.preprocessor = 'passthrough'
        return self

    def transform(self, X: PIPELINE_DATA_DTYPE) -> PIPELINE_DATA_DTYPE:
        return X

    @staticmethod
    def get_properties(dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None
                       ) -> Dict[str, Optional[Union[str, int, bool, Tuple]]]:
        return {'shortname': 'no coalescence',
                'name': 'No categorical variable coalescence',
                'handles_regression': True,
                'handles_classification': True,
                'handles_multiclass': True,
                'handles_multilabel': True,
                'handles_multioutput': True,
                'handles_sparse': True,
                'handles_dense': True,
                'input': (DENSE, SPARSE, UNSIGNED_DATA),
                'output': (INPUT,)}

    @staticmethod
    def get_hyperparameter_search_space(dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None
                                        ) -> ConfigurationSpace:
        cs = ConfigurationSpace()
        return cs
